// Globals

var AllMediaTypes = new Array("Book", "Movie", "Music", "VideoGame", "Software", "Tool", "Toy", "Gadget", "Apparel");

var DisplayedMedia = new Array(0);
var DisposedMedia = new Array(0);

var animatingResults = false;
var resultsAnimation = {duration:0, starttime:0, to:1.0, now:0.0, from:0.0, element:null, timer:null};

var flipShown = false;
var flipAnimation = {duration:0, starttime:0, to:1.0, now:0.0, from:0.0, element:null, timer:null};

var fadingSearch = false;
var searchFadeAnimation = {duration:0, starttime:0, to:1.0, now:0.0, from:0.0, element:null, timer:null};

var clickedMedium = null;


// Medium object

function Medium(url, data, width, height)
{    
    this.image = new Image(width, height);
    this.image.src = data;
    
    this.url = url;
    this.width = width;
    this.height = height;
    this.x = 0;
    this.y = 0;
    this.startX = 0;
    this.startY = 0;
    this.targetX = 0;
    this.targetY = 0;
    this.mouseOver = false;
}

Medium.prototype.toString = function()
{
    return "Medium for URL " + this.url;
}

Medium.prototype.draw = function(context)
{
    if (!this.image.complete)
        return;
    
    context.setShadow(0, 2, 3, 0.0, 0.0, 0.0, 1.0);
    context.drawImage(this.image, this.x, this.y, this.width, this.height);
    
    if (this == clickedMedium && this.mouseOver) {
        context.fillStyle = "rgba(0, 0, 0, 0.2)";
        context.fillRect(this.x, this.y, this.width, this.height);
    }    
    context.clearShadow();
}

Medium.prototype.containsPoint = function(x, y)
{
    return (x >= this.x && y >= this.y && x < this.x + this.width && y < this.y + this.height);
}


// Widget controller

function widgetStart()
{
    if (!window.libraryPlugin)
        alert("Cannot find widget plugin");
    
    document.addEventListener("mousemove", mouseMovedOverMedium, true); //???: Redundant to HTML?
    loadPreferences();
}


// Layout

function layoutShelves()
{
    layoutShelf(0, 3, shelf1Baseline());
    layoutShelf(3, 3, shelf2Baseline());
}


function layoutShelf(mediaStartIndex, mediaCount, shelfBaseline)
{
    var totalMediaWidth = 0;
    var mediaOnShelf = 0;
    
    // Calculate width of media on first shelf
    for (var mediumIndex = mediaStartIndex; mediumIndex < mediaStartIndex + mediaCount; mediumIndex++) {
        if (mediumIndex >= DisplayedMedia.length)
            break;
        
        totalMediaWidth += DisplayedMedia[mediumIndex].width;
        mediaOnShelf++;
    }
    
    // Figure out margin between media on first shelf
    var mediaMargin = (canvasWidth() - totalMediaWidth) / (mediaOnShelf + 1);
    var x = 0;
    
    for (var mediumIndex = mediaStartIndex; mediumIndex < mediaStartIndex + mediaOnShelf; mediumIndex++) {
        var medium = DisplayedMedia[mediumIndex];
        
        x += mediaMargin;
        
        medium.targetX = x;
        medium.targetY = shelfBaseline - medium.height;
        medium.y = medium.targetY;
        
        // This probably isn't the best place to put this, but we need to ensure that each medium has the proper y-value while animating (since we only interpolate along x for now).
        medium.startY = medium.targetY;
        
        x += medium.width;
    }
    
    drawShelves();
}

function drawShelves()
{
    var canvas = document.getElementById("shelf");
    var context = canvas.getContext("2d");
    
    context.clearRect(0, 0, canvasWidth(), canvasHeight());
    
    for (var mediumIndex in DisplayedMedia)
        DisplayedMedia[mediumIndex].draw(context);
    
    for (var mediumIndex in DisposedMedia)
        DisposedMedia[mediumIndex].draw(context);
    
    context.clearRect(0, shelf1Baseline(), canvasWidth(), 12); // Fake a clip (since context.clip() doesn't seem to work)
}

function disposeMedia()
{
    DisposedMedia = DisplayedMedia;
    DisplayedMedia = new Array(0);
    
    for (var mediumIndex in DisposedMedia) {
        var medium = DisposedMedia[mediumIndex];        
        var angle = (mediumIndex < 3) ? Math.PI + (Math.random() * Math.PI) : Math.random() * Math.PI;
        
        var width = canvasWidth();
        var height = canvasHeight();
        var distance = Math.sqrt(width * width + height * height) * 4; // multiply distance so animation goes faster when clearing results        
        medium.startX = medium.x;
        medium.startY = medium.y;        
        medium.targetX = Math.round(medium.startX + Math.cos(angle) * distance);
        medium.targetY = Math.round(medium.startY + Math.sin(angle) * distance);
    }
    
    startResultsAnimation();
}

function canvasWidth()
{
    return 168;
}

function canvasHeight()
{
    return 158;
}

function shelf1Baseline()
{
    return 72;
}

function shelf2Baseline()
{
    return 158;
}


// API

function doSearch(searchString)
{
    disposeMedia();
    if (!window.libraryPlugin || searchString == '')
        return;
    
    startSearchAnimation();
    window.libraryPlugin.searchWithString(searchString);
}

function hasResults(results)
{
    stopSearchAnimation();
    
    if (!results.length)
        return;
    
    DisplayedMedia = new Array(results.length);
    
    for (var resultIndex = 0; resultIndex < results.length; resultIndex++) { // for..in loop does not work for bridged RuntimeArray type
        var result = results[resultIndex];
        DisplayedMedia[resultIndex] = new Medium(result[0], result[1], parseFloat(result[2]), parseFloat(result[3]));
    }
    
    layoutShelves();
    
    if (DisplayedMedia.length > 0)
        startResultsAnimation();
}

function mediumAtPoint(x, y)
{
    // HACK
    x -= 23;
    y -= 59;
    
    for (var mediumIndex in DisplayedMedia)
        if (DisplayedMedia[mediumIndex].containsPoint(x, y))
            return DisplayedMedia[mediumIndex];
}


// Preferences (Back Side)

function loadPreferences()
{
    if (!window.widget)
        return;
    
    if (!window.widget.preferenceForKey("Searchable Media Types"))
        window.widget.setPreferenceForKey(AllMediaTypes.join(","), "Searchable Media Types");
    
    var searchableMediaTypes = window.widget.preferenceForKey("Searchable Media Types").split(",");
    var searchableMediaTypesArray = new Array();
    for (var mediumIndex in searchableMediaTypes) {
        document.getElementById("search" + searchableMediaTypes[mediumIndex] + "Checkbox").checked = true;
        searchableMediaTypesArray[mediumIndex] = searchableMediaTypes[mediumIndex];
    }
    
    if (window.libraryPlugin)
        window.libraryPlugin.setSearchableMediaTypes(window.widget.preferenceForKey("Searchable Media Types"));
}

function savePreferences()
{
    if (!window.widget)
        return;
    
    var searchableMediaTypes = new Array();
    for (var mediumIndex in AllMediaTypes)
        if (document.getElementById("search" + AllMediaTypes[mediumIndex] + "Checkbox").checked)
            searchableMediaTypes.push(AllMediaTypes[mediumIndex])
            
            window.widget.setPreferenceForKey(searchableMediaTypes.join(","), "Searchable Media Types");
    
    if (window.libraryPlugin)
        window.libraryPlugin.setSearchableMediaTypes(searchableMediaTypes.join(","));
}


// Preferences (Events)

var prefsButtonDown = false;

function showbackside(event)
{
    loadPreferences();
    
    var front = document.getElementById("front");
    var back = document.getElementById("back");
    
    if (window.widget)
        widget.prepareForTransition("ToBack");
    
    front.style.display="none";
    back.style.display="block";
    
    if (window.widget)
        setTimeout ("widget.performTransition();", 0);    
    
    document.getElementById("fliprollie").style.display = "none";
}

function enterflip(event)
{
    document.getElementById("fliprollie").style.display = "block";
}

function exitflip(event)
{
    document.getElementById("fliprollie").style.display = "none";
}

function hidePrefsMouseDown(event)
{
    prefsButtonDown = true;
    document.getElementById("hidePrefs").src = "PrefsDonePressed.png";
    document.addEventListener("mouseup", hidePrefsDocumentMouseUpHandler, true);
}

function hidePrefsDocumentMouseUpHandler(event)
{
    if (prefsButtonDown) {
        prefsButtonDown = false;
        hidePrefsMouseOver(event); // reset button image based on prefsButtonDown
    }
    
    document.removeEventListener("mouseup", hidePrefsDocumentMouseUpHandler, true);
}

function hidePrefsMouseOut(event)
{
    document.getElementById("hidePrefs").src = "PrefsDone.png";
}

function hidePrefsMouseOver(event)
{
    document.getElementById("hidePrefs").src = prefsButtonDown ? "PrefsDonePressed.png": "PrefsDone.png";    
}

function hidePrefsMouseUp(event)
{
    prefsButtonDown = false;
    
    savePreferences();
    
    var front = document.getElementById("front");
    var back = document.getElementById("back");
    
    if (window.widget)
        widget.prepareForTransition("ToFront");
    
    front.style.display = "block";
    back.style.display = "none";
    
    document.getElementById("hidePrefs").src = "PrefsDone.png";
    
    if (window.widget)
        setTimeout("widget.performTransition(); drawShelves();", 0);
}


// Animation

function animateResults()
{
    var time = (new Date).getTime();
    var T = limit_3(time - resultsAnimation.starttime, 0, resultsAnimation.duration);
    
    if (T >= resultsAnimation.duration) {
        clearInterval (resultsAnimation.timer);
        resultsAnimation.timer = null;
        resultsAnimation.now = resultsAnimation.to;
    } else {
        resultsAnimation.now = computeNextFloat (resultsAnimation.from, resultsAnimation.to, 0.5 - (0.5 * Math.cos(Math.PI * T / resultsAnimation.duration)));
    }

    // Animate displayed media
    for (var mediumIndex in DisplayedMedia) {
        var medium = DisplayedMedia[mediumIndex];
        medium.x = Math.round(medium.startX + (medium.targetX - medium.startX) * resultsAnimation.now);
        medium.y = Math.round(medium.startY + (medium.targetY - medium.startY) * resultsAnimation.now);
    }
    
    // Animate disposed media
    for (var mediumIndex in DisposedMedia) {
        var medium = DisposedMedia[mediumIndex];
        medium.x = Math.round(medium.startX + (medium.targetX - medium.startX) * resultsAnimation.now);
        medium.y = Math.round(medium.startY + (medium.targetY - medium.startY) * resultsAnimation.now);
    }
    
    if (resultsAnimation.now >= 1.0)
        stopResultsAnimation();
    
    drawShelves();
}

function startResultsAnimation()
{
    if (animatingResults) {
        resultsAnimation.now = 1.0;
        animateResults();
        stopResultsAnimation();
    }
    
    if (resultsAnimation.timer != null)
    {
        clearInterval (resultsAnimation.timer);
        resultsAnimation.timer  = null;
    }
    
    var starttime = (new Date).getTime() - 13; // set it back one frame
    
    resultsAnimation.duration = 500;
    resultsAnimation.starttime = starttime;
    resultsAnimation.timer = setInterval("animateResults();", 13);
    resultsAnimation.from = resultsAnimation.now;
    resultsAnimation.to = 1.0;
    animateResults();
    animatingResults = true;
}

function stopResultsAnimation()
{
    if (!animatingResults)
        return;
    
    if (resultsAnimation.timer != null)
    {
        clearInterval (resultsAnimation.timer);
        resultsAnimation.timer  = null;
    }
    
    resultsAnimation.now = 0.0;
    DisposedMedia = new Array(0);
    animatingResults = false;
}

function animateSearchFade()
{
    var time = (new Date).getTime();
    var T = limit_3(time - searchFadeAnimation.starttime, 0, searchFadeAnimation.duration);
    
    if (T >= searchFadeAnimation.duration) {
        clearInterval (searchFadeAnimation.timer);
        searchFadeAnimation.timer = null;
        searchFadeAnimation.now = searchFadeAnimation.to;
    } else {
        searchFadeAnimation.now = computeNextFloat (searchFadeAnimation.from, searchFadeAnimation.to, 0.5 - (0.5 * Math.cos(Math.PI * T / searchFadeAnimation.duration)));
    }
    
    searchFadeAnimation.element.style.opacity = searchFadeAnimation.now;
    searchFadeAnimation.element.style.display = searchFadeAnimation.now == 0.0 ? "none" : "block";
}

function startSearchAnimation()
{
    if (fadingSearch)
        return;
    
    // fade in the search animation
    if (searchFadeAnimation.timer != null) {
        clearInterval (searchFadeAnimation.timer);
        searchFadeAnimation.timer  = null;
    }
    
    var starttime = (new Date).getTime() - 13; // set it back one frame
    
    searchFadeAnimation.duration = 500;
    searchFadeAnimation.starttime = starttime;
    searchFadeAnimation.element = document.getElementById ("searchAnimationImage");
    searchFadeAnimation.timer = setInterval ("animateSearchFade();", 13);
    searchFadeAnimation.from = searchFadeAnimation.now;
    searchFadeAnimation.to = 1.0;
    animateSearchFade();
    fadingSearch = true;
}

function stopSearchAnimation()
{
    if (!fadingSearch) 
        return;
    
    // fade out the search animation
    if (searchFadeAnimation.timer != null)
    {
        clearInterval (searchFadeAnimation.timer);
        searchFadeAnimation.timer  = null;
    }
    
    var starttime = (new Date).getTime() - 13; // set it back one frame
    
    searchFadeAnimation.starttime = starttime;
    searchFadeAnimation.element = document.getElementById ("searchAnimationImage");
    searchFadeAnimation.timer = setInterval ("animateSearchFade();", 13);
    searchFadeAnimation.from = searchFadeAnimation.now;
    searchFadeAnimation.to = 0.0;
    animateSearchFade();
    fadingSearch = false;
}

function animateFlip()
{
    var time = (new Date).getTime();    
    var T = limit_3(time-flipAnimation.starttime, 0, flipAnimation.duration);
    
    if (T >= flipAnimation.duration) {
        clearInterval (flipAnimation.timer);
        flipAnimation.timer = null;
        flipAnimation.now = flipAnimation.to;
    } else {
        flipAnimation.now = computeNextFloat (flipAnimation.from, flipAnimation.to, 0.5 - (0.5 * Math.cos(Math.PI * T / flipAnimation.duration)));
    }
    
    flipAnimation.element.style.opacity = flipAnimation.now;
}


// Animation helpers

function limit_3 (val, min, max)
{
    return val < min ? min : (val > max ? max : val);
    //  Returns val if min < val < max
    //  Returns min if val <= min
    //  Returns max if val >= max
}

function computeNextFloat (from, to, ease)
{
    return from + (to - from) * ease;
}


// Mouse events

function mousemove (event)
{
    if (flipShown)
        return;
    
    // fade in the flip widget
    if (flipAnimation.timer != null) {
        clearInterval (flipAnimation.timer);
        flipAnimation.timer  = null;
    }
    
    var starttime = (new Date).getTime() - 13; // set it back one frame
    
    flipAnimation.duration = 500;
    flipAnimation.starttime = starttime;
    flipAnimation.element = document.getElementById ("flip");
    flipAnimation.timer = setInterval ("animateFlip();", 13);
    flipAnimation.from = flipAnimation.now;
    flipAnimation.to = 1.0;
    animateFlip();
    flipShown = true;
}

function mouseexit (event)
{
    if (!flipShown)
        return;
    
    // fade in the flip widget
    if (flipAnimation.timer != null)
    {
        clearInterval (flipAnimation.timer);
        flipAnimation.timer  = null;
    }
    
    var starttime = (new Date).getTime() - 13; // set it back one frame
    
    flipAnimation.starttime = starttime;
    flipAnimation.element = document.getElementById ("flip");
    flipAnimation.timer = setInterval ("animateFlip();", 13);
    flipAnimation.from = flipAnimation.now;
    flipAnimation.to = 0.0;
    animateFlip();
    flipShown = false;
}

function mouseDownOnShelf(event)
{        
    if (!window.libraryPlugin)
        return;
    
    var medium = mediumAtPoint(event.x, event.y);    
    if (medium)
        mouseDownOnMedium(medium, event);
}

function mouseDownOnMedium(medium, event)
{
    clickedMedium = medium;
    clickedMedium.mouseOver = true;
    drawShelves();
    
    document.addEventListener("mouseup", mouseUpOnMedium, true);
}

function mouseUpOnMedium(event)
{
    if (!window.libraryPlugin)
        return;
    
    var medium = mediumAtPoint(event.x, event.y);
    
    if (widget && medium && medium == clickedMedium)
        widget.openURL(medium.url);
    
    if (clickedMedium) {
        clickedMedium.mouseOver = false;
        clickedMedium = null;
        drawShelves();
    }
}

function mouseMovedOverMedium(event)
{
    var medium = mediumAtPoint(event.x, event.y);
    document.getElementById("shelf").style.cursor = medium ? "hand" : "auto";
    
    if (!clickedMedium)
        return;
    
    if (clickedMedium == medium && !clickedMedium.mouseOver) {
        clickedMedium.mouseOver = true;
        drawShelves();
    } else if (clickedMedium != medium && clickedMedium.mouseOver) {
        clickedMedium.mouseOver = false;
        drawShelves();
    }
}